DEFAULT_RUN_NAME = "Run"
DEFAULT_EXPERIMENT_NAME = "DanLing"
DEFAULT_EXPERIMENT_ID = "xxxxxxxxxxxxxxxx"
DEFAULT_IGNORED_KEYS_IN_HASH = {
    "iters",
    "steps",
    "epochs",
    "results",
    "index_set",
    "index",
}
